import random
import time

from pygme.game.game import Game
from pygme.game.player import Player
from pygme.snake import snake, food


class SnakeGame(Game):
    """ Defines the main Snake game loop and initialization functionality """

    def __init__(self,
                 config: dict, name: str = "Snake", difficulty: str = "normal") -> None:
        super().__init__(name, config, difficulty)
        self.board = None
        self.snake = None
        self.food_collection = None
        self.current_food = []
        self.player = Player(computer=False)

    def _validate_initialization(self, initialization_object: dict) -> None:
        """ Ensures that the given initialization_object containing parameters to run the Snake game has complete
        and valid parameters

        :param initialization_object - a dictionary containing game parameter names and their values for operation
        """
        self._validate_base(initialization_object)
        # Validate correct board dimensions
        board_width = initialization_object["board_width"]
        board_length = initialization_object["board_length"]
        # Required min coordinates for a game of snake
        required_width = self.config.get("minimum_board_width", 10)
        required_length = self.config.get("minimum_board_length", 10)
        if initialization_object["board_width"] < required_width \
                or initialization_object["board_length"] < required_length:
            raise ValueError("The Snake board must be at least {0}x{1}".format(required_length, required_width))
        if board_width != board_length:
            raise ValueError("The Snake board must be a square where width == length")

    def _initialize(self, initialization_object: dict = None) -> None:
        """ Initializes a game of Snake from the given object of game parameters or user input if one is not provided

        :param initialization_object - a dictionary containing game parameter names and their values for operation
        """
        initialization_object = self._get_user_input(initialization_object)
        board_width = initialization_object["board_width"]
        board_length = initialization_object["board_length"]
        difficulty = initialization_object["difficulty"]
        self.board = self.construct_board(board_length, board_width)
        # Pick a random point to place the snake on and a starting snake length based on chosen difficulty
        starting_x_coordinate = random.randint(2, board_length-3)
        starting_y_coordinate = random.randint(2, board_width-3)
        starting_length = {"easy": 2, "normal": 4, "hard": 8}[difficulty]
        self.snake = snake.Snake(
            x_coordinate=starting_x_coordinate, y_coordinate=starting_y_coordinate, starting_length=starting_length)
        # Start monitoring player key presses
        self.player.monitor_key_presses()
        # Create a snake food collector and generator
        self.food_collection = food.FoodCollection(grid_width=board_width, grid_length=board_length)

    def _is_game_over(self) -> bool:
        """ Checks the board to see if the game is over

        :returns True if the game is over, False otherwise
        """
        game_over = False
        current_snake_location = self.snake.current_location
        for coordinate in current_snake_location:
            # check if any coordinate is outside of the board
            # TODO: maybe only check if the head is outside for cases when snake grows into edge
            if coordinate[0] < 0 or coordinate[0] > self.board.length - 1:
                game_over = True
                break
            elif coordinate[1] < 0 or coordinate[1] > self.board.width - 1:
                game_over = True
                break
        return game_over

    def _get_food(self) -> None:
        """ Randomly generates food into the current grid """
        # TODO this should be in config
        difficulty_to_frequency_map = {
            "normal": 5, "easy": 10, "hard": 3}
        # If there is any current food on the grid that is not eaten then just return and don't generate new food
        for current_food_obj in self.current_food:
            if not current_food_obj.eaten:
                return
        # Randomly generate food based on frequency of appearance by difficulty level
        if random.randint(1, difficulty_to_frequency_map[self.difficulty]) == 1:
            generated_food_ok = False
            generated_foods = []
            # Keep generating until the food falls in a spot where it's ok
            while not generated_food_ok:
                generated_food_ok = True
                generated_foods = self.food_collection.generate()
                for generated_food in generated_foods:
                    # Don't place food in same place as Snake's head
                    if (generated_food.x_coordinate == self.snake.current_location[0][0]
                            or generated_food.y_coordinate == self.snake.current_location[0][1]):
                        # Regenerate if food is located at Snake's head
                        generated_food_ok = False
            self.current_food = generated_foods
            return
        self.current_food = []

    def _resolve_food(self, current_snake_location: list) -> None:
        """ Makes the snake eat the food and marks the food as eaten when that happens, which makes the object
        disappear from the grid; otherwise refresh the board with each uneaten piece of food

        :param current_snake_location - a list of current snake coordinates
        """
        # Check each of the current food items on the grid
        for food_obj in self.current_food:
            # If the snake's head is in the same square as the given food, then make the snake eat the food
            if (current_snake_location[0][0] == food_obj.x_coordinate
                    and current_snake_location[0][1] == food_obj.y_coordinate):
                self.snake.eat(food_obj)
                # Marking the food object as eaten will make it disappear from the grid
                food_obj.eaten = True
            # If the food still exists, show it again on the grid
            if not food_obj.eaten:
                self.board.refresh(
                    [(food_obj.x_coordinate, food_obj.y_coordinate)],
                    representation=food_obj.representation,
                    clear_board=False
                )

    def _move_snake(self, current_direction: str) -> None:
        """ Moves the snake along the grid and checks for user input entered in separate thread

        :param current_direction - the direction the snake is current traveling in with respect to the grid
        """
        # Get directional input from the user about where to go
        snake_direction = current_direction
        for key in ["left", "right", "up", "down"]:
            # Change direction only if there is a valid directional key event in the key press map
            if self.player.key_pressed_map[key]:
                snake_direction = key
                break
        # Move the snake in either the current or new direction depending on whether player pressed a key
        self.snake.move(snake_direction)

    def _finish_game(self) -> None:
        print("Game over! Hit <Enter> to play again or <q> to exit.")
        self.player.finished_game = True
        self.player.wait_for_player_to_finish()

    def run(self, initialization_object: dict = None) -> dict:
        """ Game loop that accepts player events to move the snake around the board and keeps the state of the game
        until the game finishes

        :param initialization_object - a dictionary containing game parameter names and their values for operation
        :returns a dictionary containing various metrics and their values about the game that was played
        """
        self._initialize(initialization_object)
        representation = str(self.snake)[0]
        while True:
            current_snake_location = self.snake.current_location
            self._get_food()
            self.board.refresh(current_snake_location, representation=representation)
            self._resolve_food(current_snake_location)
            self.board.print()
            print("\nHit arrow keys on your keyboard to move the snake")
            # Get the current direction of the snake
            snake_direction = self.snake.current_direction
            self._move_snake(snake_direction)
            game_over = self._is_game_over()
            if game_over:
                self._finish_game()
                break
            else:
                time.sleep(.25)
        return {}



import string

from pygme.game.board import GameBoard
from pygme.game.game import Game
from pygme.hangman import noose
from pygme.utils import dictionary


class HangmanGame(Game):
    """ Contains the necessary logic to run a game of Hangman

    Constructor parameters:

    @param config - the configuration dictionary to use for the game (should be defined in data packaged with pygme)
    @param name - the name of the game to feed to parent class
    @param difficulty - a difficulty to assume (provided by user during initialization)
    """

    def __init__(self, config: dict, name: str = "hangman", difficulty: str = "normal") -> None:
        super().__init__(name, config, difficulty)
        self.dictionary = None
        self.noose = None
        self.min_word_length = None
        self.max_word_length = None
        self.guessed_characters = set()
        self.board = None
        self.noose = None
        self.word = None

    def _validate_initialization(self, initialization_object: dict) -> None:
        """ Validates the initialization object to ensure a Hangman game can occur

        :param initialization_object - required game parameters to follow
        """
        self._validate_base(initialization_object)

    def _initialize(self, initialization_object: dict = None) -> None:
        """ Initializes the game of Hangman

        :param initialization_object - an optional dictionary containing the required game parameters to use
        """
        initialization_object = self._get_user_input(initialization_object)
        self.difficulty = initialization_object["difficulty"]
        self.dictionary = dictionary.Dictionary(self.config)
        word_sizes_by_difficulty = self.config["word_sizes_by_difficulty"][self.difficulty]
        self.min_word_length = word_sizes_by_difficulty["min_word_length"]
        self.max_word_length = word_sizes_by_difficulty["max_word_length"]
        self.board = GameBoard(self.config["board_length"], self.config["board_width"], " ")
        self.noose = noose.Noose(self.board)
        self.noose.draw()

    def _has_won(self) -> bool:
        """ Checks whether the player has won the game by checking if all the letters in the word have been guessed

        :returns True if the game has been won; False otherwise
        """
        for letter in self.word:
            if letter not in self.guessed_characters:
                return False
        return True

    def _man_died(self) -> bool:
        """ Checks whether the man has died by checking if the noose is complete

        :returns True if the man is dead; False otherwise
        """
        return self.noose.is_complete()

    def _is_game_over(self) -> bool:
        """ Checks whether the game has finished from two possible options - the player has won (guessed all letters) or
        the man has died

        :returns True if the game is over; False otherwise
        """
        if self._has_won() or self._man_died():
            return True
        return False

    def _display_word(self) -> None:
        """ Prints the current word to be guessed by hiding characters that have not been guessed yet when displaying
        the word to the user """
        if not self.guessed_characters:
            self.word.hide_all = True
        else:
            self.word.hide_all = False
            self.word.show_only = self.guessed_characters
        print("Word to guess:\n{0}\n".format(repr(self.word)))
        print("Guessed character: {0}\n".format(self.guessed_characters))

    def _display_final_status(self) -> None:
        """ Displays a message to let the user know how they did after the game finishes """
        self.word.show_only = set()
        self.word.hide_all = False
        print("\nThe word was: {0}\n".format(repr(self.word)))
        if self._has_won():
            print("Congratulations, you won!")
        else:
            print("Better luck next time!")

    def _get_guess_input(self) -> str:
        """ Accepts string input from the user to be the next guess (for monkeypatching instead of built-in input func)

        :returns the string input received from the user
        """
        return input("Enter a character to guess next:")

    def _get_guess(self) -> str:
        """ Retrieves a letter guess from the user and validates it before returning it to be used in the game

        :returns a single ASCII character to be the next letter guess in the game
        """
        guess_valid = False
        guess = None
        while not guess_valid:
            guess = self._get_guess_input()
            # Ensure the guess has not been already provided
            if guess in self.guessed_characters:
                print("You already guessed that character. Try again.")
            # There must be a guess
            elif len(guess) > 1:
                print("You can only enter one character at a time")
            # The guess must be either a lowercase or upper case letter
            elif guess not in string.ascii_letters:
                print("Guesses must be one of {0}".format(string.ascii_letters))
            else:
                guess_valid = True
        self.guessed_characters.add(guess)
        return guess

    def run(self, initialization_object: dict = None) -> dict:
        """ Runs the game of Hangman

        :param initialization_object - an optional dictionary containing the required game parameters to use
        """
        self._initialize(initialization_object)
        self.word = self.dictionary.get_random_word(min_length=self.min_word_length, max_length=self.max_word_length)
        assert len(self.word) > 1
        while not self._is_game_over():
            # Display noose and word on board
            self.board.print(join_char="")
            self._display_word()
            guess = self._get_guess()
            # Get the noose to display a new piece if the guess is wrong
            if guess not in self.word:
                self.noose.update()
            # Update the current noose on the board to be displayed to the user
            self.noose.draw()
        self.board.print(join_char="")
        self._display_final_status()
        return {}



import uuid
import threading

from curtsies import Input


class Player(object):

    def __init__(self, computer=True):
        self.player_id = uuid.uuid4()
        self.finished_game = False
        self.computer = computer
        self.keys = {"up": "'<UP>'", "right": "'<RIGHT>'", "left": "'<LEFT>'", "down": "'<DOWN>'"}
        self.event_to_key_map = {val: key for key, val in self.keys.items()}
        self.key_pressed_map = {key: False for key in self.keys}
        self.monitoring_lock = threading.Lock()
        self.thread = None
        self.winner = False

    def _detect_key_pressed(self) -> None:
        """ Detects presses of individual keys and marks those keys as pressed in common object """
        while True:
            # Check to see if any keys have been pressed
            with Input(keynames='curtsies') as input_generator:
                for e in input_generator:
                    # Iterate over the applicable keys the player can press
                    for key in self.event_to_key_map:
                        pretty_key = self.event_to_key_map[key]
                        # If the player has pressed a key that is being tracked
                        if repr(e) == key:
                            # Signal that the key has been pressed in common object
                            self.key_pressed_map = {key: False for key in self.key_pressed_map}
                            self.key_pressed_map[pretty_key] = True
                    break
            if self.finished_game:
                break

    def monitor_key_presses(self, key: str = None, how: str = "thread") -> None:
        """ Determines how to monitor for key presses by the player; whether in a separete thread or as a blocking call

        :param key - a specific key to monitor for
        :param how - how to monitor for keys pressed (thread or block)
        """
        assert how in {"thread", "block"}
        if how == "block" and not key:
            raise ValueError("A keyboard key name must be provided to block until key is pressed")
        if how == "thread":
            self.thread = threading.Thread(target=self._detect_key_pressed, args=())
            self.thread.start()
        # TODO blocking key press

    def wait_for_player_to_finish(self):
        """ Waits for the player to finish doing an action if this is being performed in another thread """
        self.thread.join()

    def __eq__(self, other):
        return self.player_id == other.player_id

    def __ne__(self, other):
        return self.player_id != other.player_id

    def __hash__(self):
        return hash(self.player_id)



from pygme.battleship import ships, player, board
from pygme.game.game import Game


class BattleshipGame(Game):

    def __init__(self,
                 config: dict, name: str = "Battleship", difficulty: str = "normal"):
        super().__init__(name, config, difficulty)
        self.players = []
        # Each player will have their own board and fleet of ships to play with
        self.boards = {}
        self.ship_fleets = {}
        # These will change depending on the current turn
        self.current_player = None
        self.other_player = None
        self.current_player_board = None
        self.other_player_board = None
        self.other_player_fleet = None
        # These maintain current ship hit stats
        self.hits_by_player = {}
        self.ships_destroyed_by_player = {}
        self.round_information = {}

    @staticmethod
    def construct_board(length: int, width: int, game_board: board.BattleshipBoard = None) -> board.BattleshipBoard:
        """ Constructs a battleship board for the game to run on

        :param length - the length of the board
        :param width - the width of the board
        :param game_board - an optional board to use instead of instantiating one
        :returns a game board to play Battleship on
        """
        if not game_board:
            game_board = board.BattleshipBoard(length, width)
        return game_board

    def _validate_initialization(self, initialization_object: dict) -> None:
        """ Validates the Battleship game initialization

        :param initialization_object - a dictionary containing starting parameters to run the game with
        """
        self._validate_base(initialization_object)
        # A default game of Battleship includes an 8x8 board
        assert (initialization_object["board_width"] >= self.config["minimum_board_width"]
                and initialization_object["board_length"] >= self.config["minimum_board_length"])

    def _initialize(self, initialization_object: dict = None) -> None:
        """ Initializes the game of battleship by choosing a human player, receiving input about the size of the
         board, creating each player's board, and placing each player's ships

        :param initialization_object - a dictionary containing starting parameters to run the game with
        """
        # Get input from the user if no initialization_object is provided
        initialization_object = self._get_user_input(initialization_object)
        board_width = initialization_object["board_width"]
        board_length = initialization_object["board_length"]
        human_player_involved = initialization_object.get("human_player_involved", True)
        self.difficulty = initialization_object["difficulty"]
        # Create boards and fleets
        board_list = [self.construct_board(board_length, board_width) for _ in range(self.number_of_players)]
        fleet_list = [ships.ShipFleet(self.config) for _ in range(self.number_of_players)]
        # Create players and associate boards, fleets, and round information objects with each one
        for idx in range(self.number_of_players):
            player_obj = player.BattleshipPlayer(board_list[idx], difficulty=self.difficulty)
            self.players.append(player_obj)
            player_obj.place_ships(
                fleet=fleet_list[idx],
            )
            self.boards[player_obj.player_id] = board_list[idx]
            self.ship_fleets[player_obj.player_id] = fleet_list[idx]
            self.hits_by_player[player_obj.player_id] = 0
            self.ships_destroyed_by_player[player_obj.player_id] = 0
            self.round_information[player_obj.player_id] = {
                "coordinate": None,
                "successful_hit": False,
                "ship_destroyed": False,
                "already_hit": False
            }
        # Assign a random player to be the human player unless specified otherwise in initialization input
        if human_player_involved:
            self._assign_human_player()

    def _is_game_over(self) -> bool:
        """ Checks whether the game has finished and determines a winner

        :returns True if the game has finished, False otherwise
        """
        players_encountered = set()
        game_over = False
        losing_player = None
        for player_in_game in self.players:
            players_encountered.add(player_in_game)
            if self.ship_fleets[player_in_game.player_id].is_destroyed():
                # Set the player with the destroyed fleet as the loser
                player_in_game.winner = False
                losing_player = player_in_game
                game_over = True
        # Set other players as the winners if the game has ended
        if game_over:
            for encountered_player in players_encountered:
                if encountered_player != losing_player:
                    encountered_player.winner = True
        return game_over

    @staticmethod
    def _get_toggle_input() -> str:
        return input("Press b to toggle between boards or a when you're ready to make a move:")

    def _display_player_status(self, enemy_board_displayed: bool) -> None:
        """ Prints out a status consisting of number of hits and ships destroyed

        :param enemy_board_displayed - whether the enemy's board is currently displayed or not
        """
        if enemy_board_displayed:
            player_id = self.current_player.player_id
        else:
            player_id = self.other_player.player_id
        successful_hits = self.hits_by_player[player_id]
        ships_destroyed = self.ships_destroyed_by_player[player_id]
        print("Hits: {0}".format(successful_hits))
        print("Ships destroyed: {0}\n".format(ships_destroyed))

    def _display_boards(self) -> None:
        """ Displays the two players' boards by providing a toggle option to switch between the current or enemy
        player's board"""
        # Start out with the other player's board displayed
        # Hide the locations of the enemy ships whenever the enemy's board is printed out
        ship_representations_to_hide = self.other_player_fleet.unique_ship_representations
        enemy_board_displayed = True
        self.other_player_board.print(include_reference=True, ignore_characters=ship_representations_to_hide)
        print("\nYou're looking at the other player's board.\n")
        # Provide the option to switch between boards or attack
        while True:
            self._display_player_status(enemy_board_displayed)
            player_input = self._get_toggle_input()
            # Print current player's board
            if player_input == "b" and enemy_board_displayed:
                self.current_player_board.print(include_reference=True)
                enemy_board_displayed = False
                print("\nYou're looking at your board.\n")
            # Print other player's board
            elif player_input == "b":
                self.other_player_board.print(include_reference=True, ignore_characters=ship_representations_to_hide)
                enemy_board_displayed = True
                print("\nYou're looking at the other player's board.\n")
            elif player_input == "a":
                break

    def _tally_hit(self, successful_hit: bool, ship_destroyed: bool) -> None:
        """ Increments ship hit status by player based on given arguments

        :param successful_hit - whether a hit was successful or not
        :param ship_destroyed - whether a hit destroyed a ship or not
        """
        if successful_hit:
            self.hits_by_player[self.current_player.player_id] += 1
        if ship_destroyed:
            self.ships_destroyed_by_player[self.current_player.player_id] += 1

    def run(self, initialization_object: dict = None) -> dict:
        """ The Battleship game loop that performs all of the resource management and user input handling to run
        the game

        :param initialization_object - a dictionary containing starting parameters to run the game with
        """
        self._initialize(initialization_object)
        already_hit = False
        # Keep running the game as long as no one has lost
        while not self._is_game_over():
            # Continue a turn if the previous hit was not for a coordinate that has not been hit before
            if not already_hit:
                # Get the players involved and resolve the current turn
                self.current_player = self._next_player()
                current_player_id = self.current_player.player_id
                other_players = self._other_players()
                assert (len(other_players) == 1)
                self.other_player = other_players[0]
                self.current_player_board = self.boards[current_player_id]
                # Get other player's information
                self.other_player_board = self.boards[self.other_player.player_id]
                self.other_player_fleet = self.ship_fleets[self.other_player.player_id]
            # Display the boards to the current player if they're human
            if not self.current_player.computer:
                self._display_boards()
            # Attack fleet
            attack_coordinate = self.current_player.guess(self.round_information[current_player_id])
            successful_hit, ship_destroyed, already_hit = self.other_player_board.attack(
                attack_coordinate, self.other_player_fleet)
            # Keep track of round status to inform next strategy
            self.round_information[current_player_id]["coordinate"] = attack_coordinate
            self.round_information[current_player_id]["successful_hit"] = successful_hit
            self.round_information[current_player_id]["ship_destroyed"] = ship_destroyed
            self.round_information[current_player_id]["already_hit"] = already_hit
            # Keep track of successful hits and ships destroyed
            self._tally_hit(successful_hit, ship_destroyed)
        self.print_result()
        return {}



import random
import uuid

from abc import ABC, abstractmethod

from pygme.utils import space


class Food(ABC):
    """ Represents snake food by keeping track of each food object's location on a Snake grid, its character
    representation in the game (how it's shown to the user on the grid), and whether the food object has been eaten
    or not.

    This class defines a common interface for different types of concrete food classes.

    Constructor arguments:

    :param food_type - the type of food
    :param representation - a string representation of the given type of food
    :param x_coordinate - the x-coordinate of the food's location on the grid
    :param y_coordinate - the y-coordinate of the food's location on the grid
    :param food_id - a unique ID that identifies the food object (generated by default)
    """

    def __init__(self,
                 food_type: str, representation: str, x_coordinate: int, y_coordinate: int,
                 food_id: uuid.UUID = uuid.uuid4()) -> None:
        self.food_type = food_type
        self.representation = representation
        self.food_id = str(food_id)
        self.x_coordinate = x_coordinate
        self.y_coordinate = y_coordinate
        self.eaten = False

    @property
    def coordinates(self) -> tuple:
        """ Retrieves the food's location on the grid
        :returns a tuple of point coordinates (x-coordinate, y-coordinate)
        """
        return self.x_coordinate, self.y_coordinate

    @abstractmethod
    def growth_value(self) -> int:
        """ Defines how much consumers of the food grow by after eating it which will depend on the subclass """
        pass

    def __repr__(self) -> str:
        """ repr() on food objects will resolve to the given string representation of each food """
        return self.representation

    def __str__(self) -> str:
        """ str() on food objects will resolve to the given string representation of each food  """
        return self.__repr__()

    def __eq__(self, other) -> bool:
        """ Two food objects are equal if their type, unique_id, and representation are the same
        :param other - a different Food object
        :returns true if the foods are the same, false otherwise
        """
        return (self.food_type == other.food_type and self.food_id == other.food_id
                and self.representation == other.representation)


class Cricket(Food):
    """ Represents the cricket type of snake food which is the most common of the types and provides the least
    amount of nutritional value for snakes

    Class attributes:

    SPAWN_WEIGHT - defines the relative frequency that crickets spawn with

    Constructor arguments:

    :param representation - a string representation for how to show crickets on the snake grid
    :param x_coordinate - the x-coordinate of the food's location on the grid
    :param y_coordinate - the y-coordinate of the food's location on the grid
    """
    SPAWN_WEIGHT = 3

    def __init__(self, representation: str = "#", x_coordinate: int = None, y_coordinate: int = None) -> None:
        super().__init__(
            food_type="crickets", x_coordinate=x_coordinate, y_coordinate=y_coordinate, representation=representation)

    @property
    def growth_value(self) -> int:
        """ Defines by how many nodes the snake grows when it consumes a cricket
        :returns a random number of nodes the snake will grow by if it eats a cricket
        """
        return random.randint(1, 2)


class Mouse(Food):
    """ Represents the mouse type of snake food which is the least common of the foods and provides the most
    nutritional value

    Class attributes:

    SPAWN_WEIGHT - defines the relative frequency that mice spawn with

    Constructor arguments:

    :param representation - a string representation for how to show mice on the snake grid
    :param x_coordinate - the x-coordinate of the food's location on the grid
    :param y_coordinate - the y-coordinate of the food's location on the grid
    """
    SPAWN_WEIGHT = 1

    def __init__(self, representation: str = "&", x_coordinate: int = None, y_coordinate: int = None):
        super().__init__(
            food_type="mouse", x_coordinate=x_coordinate, y_coordinate=y_coordinate, representation=representation)

    @property
    def growth_value(self) -> int:
        """ Defines by how many nodes the snake grows when it consumes a mouse
        :returns a random number of nodes the snake will grow by if it eats a mouse
        """
        return random.randint(2, 3)


class FoodCollection(list):
    """ Keeps a collection of different types of food objects to be used as a queue by a Snake game for randomly
    generating food objects on the grid

    Constructor arguments:

    :param grid_width - the width of the grid used by the snake game to generate food objects in
    :param grid_length - the length of the grid used by the snake game to generate food objects in
    """

    def __init__(self, grid_width: int, grid_length: int) -> None:
        super().__init__()
        self.grid_width = grid_width
        self.grid_length = grid_length
        self.length = 0
        self._refresh()

    @property
    def max_length(self) -> int:
        """ Retrieves the max length of the food collection based on the size of the grid the collection is used
        to support with random food spawns

        :returns an integer with the max number of food objects the collection can have at any time
        """
        return self.grid_width * self.grid_length

    def _reset(self) -> None:
        """ Resets the current collection by emptying it and making room for new food objects up to the max length
        allowed
        """
        # Empty the current food collection
        self[:] = []
        # Create space for populating it with new food items
        self.extend([None] * self.max_length)
        self.length = self.max_length

    def _refresh(self) -> None:
        """ Provides a new bash of food objects in the current food collection based on how often each food
        object should appear when chosen randomly
        """
        # Empty the collection and make room for new food objects
        self._reset()
        # Enumerate the types of food to choose from and their relative weights of appearance frequency
        eligible_types = [Mouse, Cricket]
        weights = [Mouse.SPAWN_WEIGHT, Cricket.SPAWN_WEIGHT]
        assert len(eligible_types) == len(weights)
        # Generate max_length number of food objects randomly in a stratified fashion based on their weights
        random_type_choices = random.choices(eligible_types, weights=weights, k=self.max_length)
        for idx, food_type in enumerate(random_type_choices):
            # Assign random coordinates to each food item for the game to use for grid placement
            random_coordinates = space.get_coordinates_between_limits(self.grid_width, self.grid_length)
            self[idx] = food_type(x_coordinate=random_coordinates[0], y_coordinate=random_coordinates[1])

    def generate(self, count: int = 1) -> list:
        """ Generates the given count of random food objects from the collection and refreshes the collection
        with new objects if the collection is empty

        :param count - the number of food objects to generate
        :returns a list of randomly chosen food objects
        """
        assert count > 0
        return_foods = []
        for _ in range(count):
            # Refresh the current collection with new food objects if the collection is empty
            if self.length == 0:
                self._refresh()
            # Get last object from the food collection and remove it from the collection
            return_foods.append(self.pop())
            self.length -= 1
        return return_foods



import pytest

import pygme.__main__ as cli


def test_load_config():
    """ Tests pygme.__main__._load_config function """
    config_dict = cli._load_config()
    assert isinstance(config_dict, dict) and len(config_dict) > 0
    required_games_to_check = [key for key in cli.SUPPORTED_GAMES]
    for game in required_games_to_check:
        assert game in config_dict


def test_validate_config():
    """ Tests pygme.__main__._validate_config function """
    dummy_config = {"some_other_game": {}}
    with pytest.raises(ValueError):
        cli._validate_config(dummy_config)
    actual_config = cli._load_config()
    cli._validate_config(actual_config)






def validate_user_input(input_variable: str, value_received: any, expected_type: type) -> any:
    """ Validates the given input variable value by checking if it can be converted to the expected type and if
    so, converts the variable to that type

    :param input_variable - the name of the variable to validate
    :param value_received - the value of the variable to validate
    :param expected_type an expected type that the variable should have
    returns the same variable as is or converted to the expected type that it was validated to be
    """
    expected_type_name = expected_type.__name__
    try:
        return expected_type(value_received)
    except ValueError:
        raise ValueError("Your input for {0} cannot be represented as {1}".format(input_variable, expected_type_name))


def validate_out_of_possible_options(user_input: str, options: set):
    """ Raises an exception is the given input in not contained within the given input options

    :param user_input - an input to validate
    :param options - the options that the input must be part of
    """
    if user_input not in options:
        raise ValueError("Your input of {0} is not one of {1}".format(user_input, options))


def validate_grid_index(grid_length: int, grid_width: int, x_coordinate: int, y_coordinate: int) -> bool:
    # TODO already implemented in space.py as are_coordinates_between_limits
    if x_coordinate < 0 or x_coordinate >= grid_length or y_coordinate < 0 or y_coordinate >= grid_width:
        return False
    return True





def resolve_movement(x_coordinate: int, y_coordinate: int, direction: str) -> tuple:
    """ Given current coordinates and a movement around a 2D grid, this function will return new coordinates that
    reflect that movement in 2 dimensional space

    :param x_coordinate - x-coordinate identifying the starting point on a 2D grid
    :param y_coordinate - y-coordinate identifying the starting point on a 2D grid
    :param direction - one of four possible directions to move in
    :returns a tuple of integers representing a new point (x-coordinate, y-coordinate)
    """
    assert direction in {'up', 'right', 'down', 'left'}
    if direction == "up":
        return x_coordinate, y_coordinate - 1
    elif direction == "right":
        return x_coordinate + 1, y_coordinate
    elif direction == "down":
        return x_coordinate, y_coordinate + 1
    elif direction == "left":
        return x_coordinate - 1, y_coordinate



from abc import ABC, abstractmethod

from pygme.game import board
from pygme.utils import space


class Strategy(ABC):
    """ Provides an interface for different game strategies

    Constructor parameters:

    @param game_board - a board on which the game is being played
    @param name - the name of the strategy
    """
    def __init__(self, game_board: board.GameBoard, name: str) -> None:
        self.game_board = game_board
        self.name = name
        self.previous_coordinate_attacked = None
        self.previous_attack_hit = False
        self.previous_attack_sank_ship = False
        self.already_guessed = set()

    def _process_information(self, information: dict) -> None:
        """ Retrieves the required pieces information needed to make a guess

        @param information - information about the previous round to base the guessing decision on
        """
        self.previous_coordinate_attacked = information["coordinate"]
        self.previous_attack_hit = information["successful_hit"]
        self.previous_attack_sank_ship = information["ship_destroyed"]

    def _random_guess(self) -> tuple:
        """ Guess randomly from the list of grid coordinates not already guessed

        :returns a tuple of X and Y coordinates
        """
        return space.get_coordinates_between_limits(
            self.game_board.width, self.game_board.length, exclusion_set=self.already_guessed)

    @abstractmethod
    def guess(self, information: dict) -> tuple:
        pass



from pygme.battleship.ships import Ship, ShipFleet
from pygme.game import board

from pygme.utils import space


class BattleshipBoard(board.GameBoard):
    """ Implements specific battleship board functionality

    Constructor parameters:

    :param length - the length of the board to create
    :param width - the width of the board to create
    :param empty_square - how to represent empty squares on the board
    """
    def __init__(self, length: int, width: int, empty_square: str = "_") -> None:
        super().__init__(length, width, empty_square)

    def place_ship(self, ship: Ship) -> None:
        """ Refreshes the board with the coordinates of a given ship """
        self.refresh(ship.coordinates, ship.representation, clear_board=False)

    def attack(self, coordinate: tuple, fleet: ShipFleet) -> tuple:
        """ Accepts an attack from a player to a given fleet on the board by transmitting the attack to the fleet and
        refreshing the board with ship coordinates (which should include newly damaged segments of ships)

        :param coordinate - a tuple of x and y coordinates to attack on the board
        :param fleet - a fleet residing on the board to attack
        :returns a tuple containing booleans denoting whether a ship was successfully hit and destroyed or whether a
            ship has already been hit at the given coordinates
        """
        assert space.are_coordinates_between_limits(coordinate, self.width, self.length)
        successful_hit, ship_destroyed, already_hit = False, False, False
        if fleet.ship_is_hit(coordinate):
            already_hit = True
        else:
            # Attack the fleet
            successful_hit, ship_destroyed, representation = fleet.accept_attack(coordinate)
            # If the attack was successful, then refresh the board with the new damaged representation of the ship
            self.refresh([coordinate], representation="@", clear_board=False)
            if successful_hit:
                self.refresh([coordinate], representation, clear_board=False)
        # Emit whether the attack was successful and if a ship was destroyed
        return successful_hit, ship_destroyed, already_hit



from pygme.utils.display import clear_console
from pygme.utils.validation import validate_grid_index
from pygme.utils.space import are_coordinates_between_limits


class GameBoard(object):
    """ Represents a base board to play a game on which may be extended by more specific types of boards

    Constructor arguments:

    :param length - the length of the board to create
    :param width - the width of the board to create
    :param empty_square - how to represent empty squares on the board
    """
    def __init__(self, length: int, width: int, empty_square: str = "_") -> None:
        assert length > 0 and width > 0
        self.length = length
        self.width = width
        self.empty_square = empty_square
        self.board = []
        self._create_board()

    def is_square_clear(self, coordinate: tuple) -> bool:
        """ Returns True if the square at a given coordinate is empty; False if it's not empty

        :param coordinate - an x,y coordinate to check
        :returns whether the given coordinate is empty in the board
        """
        assert are_coordinates_between_limits(coordinate, self.width, self.length)
        if self.board[coordinate[0]][coordinate[1]] == self.empty_square:
            return True
        return False

    def _create_board(self) -> None:
        """ Creates an empty 2D list with the given board dimensions"""
        for i in range(self.length):
            self.board.append([self.empty_square for _ in range(self.width)])

    def print(self, include_reference: bool = False, ignore_characters: set = None, join_char: str = " ") -> None:
        """ Prints out the board to stdout

        :param include_reference - whether to include grid references when printing the board out in the console
        :param ignore_characters - a collection of characters to replace with empty characters
        :param join_char - how to join a row list together to produce a row on the terminl screen
        """
        # Clear the terminal
        clear_console()
        if not ignore_characters:
            ignore_characters = set()
        # Add an index before each column if applicable
        if include_reference:
            header_items = [str(x) for x in range(self.length)]
            header = "    "
            for item in header_items:
                if len(item) == 1:
                    header += item + "   "
                else:
                    header += item + "  "
            print(header)
        # Print the board
        for i in range(self.width):
            # Print an index before each row if applicable
            header = ""
            space = ""
            if include_reference:
                header = "{0}".format(i)
                if len(header) == 1:
                    space = " " * 3
                else:
                    space = " " * 2
                row_string = ""
                for square in range(self.length):
                    column_spacing = "   "
                    if self.board[square][i] and self.board[square][i] not in ignore_characters:
                        row_string += self.board[square][i]
                    else:
                        row_string += self.empty_square
                    row_string += column_spacing
                print(header + space + row_string + "\n")
            else:
                row_string = join_char.join([self.empty_square if (not self.board[square][i]
                                                                   or self.board[square][i] in ignore_characters)
                                             else self.board[square][i] for square in range(self.length)])
                print(row_string)

    def clear(self) -> None:
        """ Clears the current board by replacing every square with the given empty square character """
        for i in range(self.width):
            for j in range(self.length):
                self.board[j][i] = self.empty_square

    def refresh(self,
                coordinates: list, representation: str, clear_board: bool = True) -> None:
        """ Refreshes the board by adding the given representation character to the given coordinates

        Example: representation = '*' at coordinates [(0, 1), (2, 1)] on a 3x3 board will result in the following:
        _ _ _
        * _ *
        _ _ _

        :param coordinates - a list of coordinate tuples to update
        :param representation - the character to be placed in the given coordinates
        :param clear_board - whether to first clear the current board before placing the new characters or not
        """
        # Clear the current board first if the provided argument is true
        if clear_board:
            self.clear()
        for coordinate_tuple in coordinates:
            x_coordinate, y_coordinate = coordinate_tuple[0], coordinate_tuple[1]
            # Only refresh the board with the coordinate if the coordinate is valid
            if validate_grid_index(self.length, self.width, x_coordinate, y_coordinate):
                # Refresh the board
                self.board[x_coordinate][y_coordinate] = representation

    def __repr__(self):
        return "GameBoard ({0} by {1})".format(self.length, self.width)

    def __str__(self):
        return self.__repr__()



from pygme.game import strategy
from pygme.battleship import board
from pygme.utils import space


class RandomGuessStrategy(strategy.Strategy):

    def __init__(self, game_board: board.BattleshipBoard, name: str = "Random Guess"):
        super().__init__(game_board, name)
        self.targets = []

    def guess(self, information: dict) -> tuple:
        """ Guesses a coordinate to attack purely at random

        :param information - the information about the last round (not used)
        :returns a coordinate tuple to attack next
        """
        guess = self._random_guess()
        self.already_guessed.add(guess)
        return guess


class HuntTargetStrategy(strategy.Strategy):
    """ Hunt/Target guessing strategy

    - If the last hit was successful and a ship has yet to be destroyed, add adjacent squares to stack
    - If the last hit was successful and a ship was sunk, empty the current stack
    - Always guess the last target in the stack (Target)
    - If the stack is empty, return a random guess (Hunt)

    Constructor parameters:

    @param game_board - a Battleship board on which the game is being played
    @param name - the name of the strategy
    """
    def __init__(self, game_board: board.BattleshipBoard, name: str = "Hunt and Target"):
        super().__init__(game_board, name)
        self.targets = []

    def _add_coordinates_to_targets(self, coordinates: list) -> None:
        """ Adds the given list of coordinates to the stack of target coordinates to attack

        :param coordinates - a list of coordinates to target
        """
        for coordinate in coordinates:
            if coordinate not in self.targets:
                self.targets.append(coordinate)

    def guess(self, information: dict) -> tuple:
        """ Guesses a coordinate to attack based on previous information according to the Hunt/Target strategy

        :param information - the information about the last round to base the guessing decision on
        :returns a coordinate tuple to attack next
        """
        self._process_information(information)
        guess = tuple()
        valid_guess = False
        while not valid_guess:
            if not self.previous_coordinate_attacked:
                break
            # Target mode: previous attack resulted in hit but did not destroy ship so add adjacent squares to stack
            if (self.previous_attack_hit and not self.previous_attack_sank_ship
                    and self.previous_coordinate_attacked not in self.already_guessed):
                coordinates = space.get_adjacent_coordinates(
                    self.previous_coordinate_attacked, self.game_board.width, self.game_board.length)
                self._add_coordinates_to_targets(coordinates)
            # Previous attack sank ship so empty the list of targets and resume hunt mode with random guess
            elif self.previous_attack_sank_ship:
                self.targets = []
            # The last target will always be first guess if there are targets available; otherwise it will be random
            if len(self.targets) > 0:
                guess = self.targets.pop()
            else:
                guess = self._random_guess()
            if guess not in self.already_guessed:
                valid_guess = True
            self.already_guessed.add(self.previous_coordinate_attacked)
        return guess


class ProbabilityDensityStrategy(strategy.Strategy):

    def __init__(self, game_board: board.BattleshipBoard, name: str = "Random Guess"):
        super().__init__(game_board, name)
        self.targets = []

    def guess(self, information: dict) -> tuple:
        """ Guesses a coordinate to attack based on the probability densities of each ship

        :param information - the information about the last round
        :returns a coordinate tuple to attack next
        """
        raise NotImplementedError



from abc import ABC

from pygme.game.board import GameBoard


class Part(ABC):
    """ A base Part interface to be inherited by different types of concrete parts

    Constructor parameters:

    @param part - the name of the part
    @param part_char_map - how the part will be represented to the user on a grid
    """
    def __init__(self, part: str, part_char_map: dict):
        self.part = part
        self.part_char_map = part_char_map

    def __repr__(self):
        return self.part_char_map[self.part]

    def __str__(self):
        return self.__repr__()


class BodyPart(Part):
    """ Represents part of the man in Hangman

    Constructor parameters:

    @param part - the name of the part
    """

    PART_CHAR_MAP = {
        "head": "O",
        "left_arm": "-",
        "right_arm": "-",
        "left_leg": "/",
        "right_leg": "\\",
        "body": "|"
    }

    def __init__(self, part: str):
        super().__init__(part, self.PART_CHAR_MAP)


class NoosePart(Part):
    """ Represents part of noose in Hangman

    Constructor parameters:

    @param part - the name of the part
    """

    PART_CHAR_MAP = {
        "base": "_",
        "top": "_",
        "pole": "|",
        "rope": "|"
    }

    def __init__(self, part: str):
        super().__init__(part, self.PART_CHAR_MAP)


class Noose(object):
    """ Combines noose and body parts together to make up the Hangman noose that is displayed to the user

    Constructor parameters:

    @param game_board - the board to draw the noose on
    """
    def __init__(self, game_board: GameBoard) -> None:
        self.game_board = game_board
        self.noose_components = [
            # The base of the noose and rope
            {"x_index": 0, "y_index": 5, "part": NoosePart("base"), "displayed": True},
            {"x_index": 1, "y_index": 5, "part": NoosePart("pole"), "displayed": True},
            {"x_index": 2, "y_index": 5, "part": NoosePart("base"), "displayed": True},
            {"x_index": 1, "y_index": 4, "part": NoosePart("pole"), "displayed": True},
            {"x_index": 1, "y_index": 3, "part": NoosePart("pole"), "displayed": True},
            {"x_index": 1, "y_index": 2, "part": NoosePart("pole"), "displayed": True},
            {"x_index": 1, "y_index": 1, "part": NoosePart("pole"), "displayed": True},
            {"x_index": 2, "y_index": 0, "part": NoosePart("top"), "displayed": True},
            {"x_index": 3, "y_index": 0, "part": NoosePart("top"), "displayed": True},
            {"x_index": 4, "y_index": 1, "part": NoosePart("rope"), "displayed": True},
            # The body
            {"x_index": 4, "y_index": 2, "part": BodyPart("head"), "displayed": False},
            {"x_index": 4, "y_index": 3, "part": BodyPart("body"), "displayed": False},
            {"x_index": 3, "y_index": 3, "part": BodyPart("left_arm"), "displayed": False},
            {"x_index": 5, "y_index": 3, "part": BodyPart("right_arm"), "displayed": False},
            {"x_index": 3, "y_index": 4, "part": BodyPart("left_leg"), "displayed": False},
            {"x_index": 5, "y_index": 4, "part": BodyPart("right_leg"), "displayed": False},
        ]
        self.next_piece = self.get_last_displayed() + 1

    def draw(self) -> None:
        """ Refreshes the associated grid with the latest noose representation """
        self.game_board.clear()
        for component in self.noose_components:
            if not component["displayed"]:
                continue
            part = component["part"]
            coordinate = (component["x_index"], component["y_index"])
            self.game_board.refresh([coordinate], repr(part), clear_board=False)

    def is_complete(self) -> bool:
        """ Checks whether all components have been displayed, meaning the noose is complete

        :returns True if the entire noose has been build during the course of the Hangman game; False otherwise
        """
        complete = True
        for component in self.noose_components:
            if not component["displayed"]:
                complete = False
        return complete

    def get_last_displayed(self) -> int:
        """ Retrieves the last component of the Hangman noose that is currently displayed to the user

        :returns an index to the last displayed component in the noose_components class attribute
        """
        complete_idx = len(self.noose_components) - 1
        for idx, component in enumerate(self.noose_components):
            if not component["displayed"]:
                complete_idx = idx - 1
                break
        return complete_idx

    def update(self) -> None:
        """ Updates the noose by displaying a new component """
        if not self.is_complete():
            component_to_update = self.noose_components[self.next_piece]
            component_to_update["displayed"] = True
            self.next_piece += 1



from pygme.game import movement


class Node(object):
    """ Represents each node in the body of the snake """

    def __init__(self, x_coordinate: int, y_coordinate: int, direction: str,
                 representation: str = '*', next_node=None, prev_node=None) -> None:
        self.x_coordinate = x_coordinate
        self.y_coordinate = y_coordinate
        self.direction = direction
        self.representation = representation
        self.next_node = next_node
        self.prev_node = prev_node

    def __repr__(self):
        return self.representation

    def __str__(self):
        return self.__repr__()


class Body(object):
    """ Represents the body of the snake; Defines common functionality for growing and moving """

    def __init__(self, x_coordinate: int, y_coordinate: int, length: int = 1, direction: str = "left") -> None:
        self.x_coordinate = x_coordinate
        self.y_coordinate = y_coordinate
        self.head = Node(x_coordinate, y_coordinate, direction)
        self.tail = self.head
        self.length = 1
        self.direction = direction
        # Grow the body to the desired initial length
        if length > 1:
            self.grow(length - 1)

    @property
    def coordinates(self) -> list:
        coordinates_list = []
        tmp_node = self.head
        while tmp_node:
            coordinates_list.append((tmp_node.x_coordinate, tmp_node.y_coordinate))
            tmp_node = tmp_node.next_node
        return coordinates_list

    def change_direction(self, new_direction: str) -> None:
        """ Changes the current body's direction of movement

        :param new_direction - one of four possible directions to move the body in
        """
        opposite_direction = {"left": "right", "right": "left", "down": "up", "up": "down"}[self.direction]
        # Only change the current direction if it's not opposite of the current direction
        if new_direction != opposite_direction:
            self.direction = new_direction

    @staticmethod
    def _create_new_node(direction: str, prev_node: Node):
        x_coordinate = None
        y_coordinate = None
        if direction == "left":
            x_coordinate = prev_node.x_coordinate + 1
            y_coordinate = prev_node.y_coordinate
        elif direction == "right":
            x_coordinate = prev_node.x_coordinate - 1
            y_coordinate = prev_node.y_coordinate
        elif direction == "up":
            x_coordinate = prev_node.x_coordinate
            y_coordinate = prev_node.y_coordinate + 1
        elif direction == "down":
            x_coordinate = prev_node.x_coordinate
            y_coordinate = prev_node.y_coordinate - 1
        new_node = Node(x_coordinate, y_coordinate, direction, prev_node=prev_node)
        return new_node

    def grow(self, by: int = 1) -> None:
        """ Grows the current body by the specified amount of nodes

        :param by - the number of nodes to grow the body by
        """
        assert by > 0
        new_node_chain = None
        last_node = None
        # Create a new chain of nodes to append to the end of the body
        for new_node_num in range(by):
            # Create the first node in the chain
            if new_node_num == 0:
                new_node_chain = self._create_new_node(direction=self.tail.direction, prev_node=self.tail)
                last_node = new_node_chain
            # Create all the other nodes after the first one and add them to the chain
            else:
                tmp_node = self._create_new_node(direction=last_node.direction, prev_node=last_node)
                last_node.next_node = tmp_node
                last_node = tmp_node
        # Integrate the new chain as part of the current body
        self.tail.next_node = new_node_chain
        self.tail = last_node
        self.length += by

    def slither(self, speed: float = 1.0) -> None:
        """ Moves the body in its current direction of movement

        :param speed - the speed to move the body in
        """
        # Each node will assign its direction and coordinate to the next node
        tmp_node = self.head.next_node
        tmp_x_coordinate = self.head.x_coordinate
        tmp_y_coordinate = self.head.y_coordinate
        while tmp_node:
            new_x_coordinate = tmp_node.x_coordinate
            new_y_coordinate = tmp_node.y_coordinate
            tmp_node.x_coordinate = tmp_x_coordinate
            tmp_node.y_coordinate = tmp_y_coordinate
            tmp_node = tmp_node.next_node
            tmp_x_coordinate = new_x_coordinate
            tmp_y_coordinate = new_y_coordinate
        # Once all other nodes have moved, the head moves to new location
        self.head.x_coordinate, self.head.y_coordinate = movement.resolve_movement(
            self.head.x_coordinate, self.head.y_coordinate, self.direction)

    def __repr__(self):
        node_list = []
        tmp_node = self.head
        while tmp_node:
            node_list.append(str(tmp_node))
            tmp_node = tmp_node.next_node
        return "-".join(node_list)

    def __str__(self):
        return self.__repr__()


class Snake(object):
    """ Represents the snake itself

    Constructor arguments:

    :param x_coordinate - the starting x-coordinate of the snake's head
    :param y_coordinate - the starting y-coordinate of the snake's head
    :param starting_length - the snake's starting node length
    :param starting_direction - the starting direction of movement that the snake will immediately head in
    """
    def __init__(self,
                 x_coordinate: int, y_coordinate: int,
                 starting_length: int = 1, starting_direction: str = "left") -> None:
        self.body = Body(x_coordinate, y_coordinate, length=starting_length, direction=starting_direction)

    @property
    def current_location(self) -> list:
        """ Retrieves a list of coordinate tuples representing the snake's location on a snake grid

        Example: [(0,1), (1,1)] is a 2 node long snake and on a 3 x 3 grid will look like this:

        _ _ _
        * * _
        _ _ _

        :returns a list of coordinate tuples
        """
        return self.body.coordinates

    @property
    def current_direction(self) -> str:
        """ Retrieves the snake's current direction of movement

        :returns a direction string (left, up, right, down)
        """
        return self.body.direction

    def eat(self, food_to_eat: any) -> None:
        """ Eats a given food and makes the snake grow by the food's growth value

        :param food_to_eat - a food object as defined in food.py
        """
        self.body.grow(food_to_eat.growth_value)

    def move(self, new_direction: str) -> None:
        """ Moves the snake by changing the direction to the new direction and altering the corresponding
        coordinates based on that direction of movement

        :param new_direction - the direction of movement (left, up, right, down)
        """
        # Change direction
        self.body.change_direction(new_direction)
        # Move
        self.body.slither()

    def __repr__(self):
        return self.body.__repr__()

    def __str__(self):
        return self.body.__str__()



import pathlib

from setuptools import setup, find_packages


# The directory containing this file
HERE = pathlib.Path(__file__).parent

# The text of the README file
README = (HERE / "README.md").read_text()


setup(
    description='Lightweight computer games written in Python',
    long_description=README,
    long_description_content_type="text/markdown",
    url="https://github.com/adaros92/pygme",
    version='0.3.0',
    install_requires=['curtsies'],
    tests_require=['pytest', 'pytest-cov', 'tox', 'Random-Word'],
    license="MIT",
    classifiers=[
            "License :: OSI Approved :: MIT License",
            "Programming Language :: Python :: 3",
            "Programming Language :: Python :: 3.7",
        ],
    packages=find_packages(exclude=("test",)),
    name='pygme',
    python_requires='>=3.6',
    package_data={
            'pygme': ['data/*']
        },
    entry_points={
        'console_scripts': [
                'pygme = pygme.__main__:main'
            ]
    }
)



from pygme.battleship import ships, board, strategy
from pygme.game import player
from pygme.utils import space


class BattleshipPlayer(player.Player):

    DIFFICULTY_TO_STRATEGY_MAP = {
        "easy": strategy.RandomGuessStrategy,
        "normal": strategy.HuntTargetStrategy,
        "hard": strategy.ProbabilityDensityStrategy
    }

    def __init__(self, game_board: board.BattleshipBoard, computer=True, difficulty: str = "normal"):
        super().__init__(computer=computer)
        self.strategy = self.DIFFICULTY_TO_STRATEGY_MAP[difficulty](game_board)
        self.random_strategy = strategy.RandomGuessStrategy(game_board)
        self.game_board = game_board

    def _get_coordinate_from_input(self) -> tuple:
        """ Accepts user attack coordinate input

        :returns a tuple containing the X and Y coordinates received as user input
        """
        x_coordinate_guess = int(input("Enter the x-coordinate of the square to attack:"))
        y_coordinate_guess = int(input("Enter the y-coordinate of the square to attach:"))
        return x_coordinate_guess, y_coordinate_guess

    def _human_guess(self) -> tuple:
        """ Accept an attack guess from a human player as raw input

        :returns a tuple containing the X and Y coordinates of the square to attack
        """
        input_valid = False
        coordinate = None
        while not input_valid:
            try:
                coordinate = self._get_coordinate_from_input()
            except ValueError:
                print("Individual coordinate components must be single integers")
                continue
            if space.are_coordinates_between_limits(coordinate, self.game_board.width, self.game_board.length):
                input_valid = True
        return coordinate

    def _computer_guess(self, information: dict) -> tuple:
        """ Run computer attack guessing algorithm

        :param information - information about the previous round to inform the current round's strategy
        :returns a tuple containing the X and Y coordinates of the square to attack
        """
        # If the previous round coordinate was for a square that was already hit then return a guess at random
        # Also return random guess if the previous guess did not contain any info (such as when first guessing)
        if information["already_hit"] or not information["coordinate"]:
            return self.random_strategy.guess(information)
        return self.strategy.guess(information)

    def guess(self, information: dict) -> tuple:
        """ Each player will guess the other's ships' locations and attack them one at a time

        :param information - information about the previous round to inform the current round's strategy
        :returns a tuple containing the X and Y coordinates of the square to attack
        """
        # Accept input from player if they're not the computer
        if not self.computer:
            coordinate = self._human_guess()
        # Follow guessing algorithm if computer
        else:
            coordinate = self._computer_guess(information)
        return coordinate

    # TODO: move this to utils or game class
    @staticmethod
    def _validate_ship_placement(coordinates: list, game_board: board.BattleshipBoard) -> None:
        """ Validates that the given coordinates for ship placement on the given board are valid

        :param coordinates - a list of x and y coordinate tuples to check the validity of
        :param game_board - the board object to place the given coordinates in
        """
        for coordinate in coordinates:
            if len(coordinate) != 2:
                raise ValueError("A coordinate must be a tuple consisting of X and Y integers")
            if not space.are_coordinates_between_limits(coordinate, game_board.width, game_board.length):
                raise ValueError("The location of the ship is not within the boundaries of the grid")
            if not game_board.is_square_clear(coordinate):
                raise ValueError("The location of the ship overlaps another ship")

    def _place_ship(self, ship: ships.Ship) -> None:
        """ Place a given ship on the given board

        Placement works by picking a starting coordinate and generating contiguous slots on the board from that starting
        location in one of two possible directions - right or down

        :param ship - the ship to place on the board
        """
        ship_placement_ok = False
        ship_directions = {"right", "down"}
        # Keep trying to place the ship until it doesn't overlap with anything that's currently on the board
        while not ship_placement_ok:
            try:
                # Choose a random location and starting direction
                board_width, board_length = self.game_board.width, self.game_board.length
                starting_coordinate = space.get_coordinates_between_limits(board_width, board_length)
                direction = space.get_random_directions(list(ship_directions), 1)[0]
                # Retrieve the coordinates based random starting location of ships
                coordinates = space.get_contiguous_coordinates(starting_coordinate, direction, ship.size)
                # Raise exception if any of the ships go beyond the board or if they overlap others
                self._validate_ship_placement(coordinates, self.game_board)
                # Let ship object know that it is placed on a grid
                ship.place_ship(coordinates)
                ship_placement_ok = True
                # Update the board with the new ship
                self.game_board.place_ship(ship)
            except ValueError as e:
                continue

    def place_ships(self, fleet: ships.ShipFleet) -> None:
        """ Places each ship in the given fleet on the given board

        :param fleet - the fleet containing ships to be placed on the board
        """
        for ship_name, ship in fleet.items():
            self._place_ship(ship)



import argparse
import json
import os
import pkg_resources

from pygme.snake.game import SnakeGame
from pygme.battleship.game import BattleshipGame
from pygme.hangman.game import HangmanGame

SUPPORTED_GAMES = {"snake": SnakeGame, "battleship": BattleshipGame, "hangman": HangmanGame}


def _load_config() -> dict:
    """ Gets a config object from a JSON file packaged up with the distribution

    :returns a dictionary with key-value config parameters
    """
    directory_path = pkg_resources.resource_filename('pygme', 'data/')
    full_path = os.path.join(directory_path, "config.json")
    with open(full_path, "r") as f:
        config = json.load(f)
    return config


def _validate_config(config: dict) -> None:
    """ Validates a given config dictionary by raising exceptions when something is off before starting a game

    :param config - a configuration dictionary to validate
    """
    # Check to ensure that each game has an entry in the package's config
    for game_name in SUPPORTED_GAMES:
        if game_name not in config:
            raise ValueError("The provided config in pygme/data/config.json must have settings for {0}".format(
                game_name))


def _parse_args():
    """ Parse arguments passed in via the CLI
    :returns a parsed argument object from argparse
    """
    parser = argparse.ArgumentParser()
    # Accept the operation to perform
    parser.add_argument(
        'game', choices=[game_name for game_name, game_obj in SUPPORTED_GAMES.items() if game_obj],
        help="choose the game to play")
    return parser.parse_args()


def main():
    # Load configuration and validate it
    config = _load_config()
    _validate_config(config)
    # Get arguments from CLI
    args = _parse_args()
    # Instantiate and run corresponding game class with provided config
    game_class = SUPPORTED_GAMES[args.game]
    if not game_class:
        raise RuntimeError("Game is unavailable. Check back later!")
    game_config = config[args.game]
    # Assume a default dictionary if none provided for word games
    game_config["dictionary_filename"] = game_config.get("dictionary_filename", "dictionary.txt")
    game_object = game_class(config=game_config)
    game_object.run()


if __name__ == "__main__":
    main()



import os
import pkg_resources
import random
import sys


class Word(object):
    """ Represents a word in a dictionary. Used to provide supplements to existing str class by not taking into
    account capitalization when comparing words and hiding certain characters when returning the word to a caller.

    Constructor parameters:

    @param word - a regular string word
    @param show_only - an optional set containing characters to exclude from the word when repr or str are called
    """

    def __init__(self, word: str, show_only: set = None):
        super().__init__()
        if " " in word:
            raise ValueError("The given word {0} contains whitespace".format(word))
        self.word = word.lower()
        self.show_only = show_only
        self.hide_all = False
        if not show_only:
            self.show_only = set()

    def __len__(self):
        return len(self.word)

    def __contains__(self, letter: str) -> bool:
        """ Checks whether the given letter is within the word without taking capitalization into account

        For example:
        "a" in action == True
        "A" in action == True
        "p" in action == False

        :param letter - the letter to check
        :returns True if the letter is in the word, False otherwise
        """
        return letter.lower() in self.word

    def __repr__(self) -> str:
        """ When repr() is called on the string, exclude characters not in the self.show_only set if it's nonempty

        :returns either the string as is or the string with excluded
        """
        return_str = self.word
        if self.show_only or self.hide_all:
            return_str = ""
            # Only show characters that are in the given show_only set
            for char in self.word:
                if char.lower() in self.show_only or char.upper() in self.show_only:
                    return_str += char
                else:
                    return_str += "_"
        return return_str

    def __str__(self) -> str:
        return self.__repr__()

    def __setattr__(self, key: str, value: any) -> None:
        """ Overrides the set attribute functionality to always take the lowercase of a word

        For example:
        my_word = Word("hello")
        my_word.word = "Hi"
        my_word.word == "hi" # True

        :param key - the object attribute to set
        :param value - the value to assign to the attribute
        """
        # Any attribute other than word will be set normally and word will always be taken as lowercase
        self.__dict__[key] = value
        if key == "word":
            self.__dict__[key] = value.lower()

    def __iter__(self):
        yield from self.word


class Dictionary(object):
    """ Represents a dictionary to get words from for word games

    Constructor arguments:

    :param config - a configuration dictionary containing the file packaged up with pygme containing all words
    """
    def __init__(self, config: dict) -> None:
        self.dictionary_filename = config["dictionary_filename"]
        self.words = []
        self._load_dictionary()

    def _load_dictionary(self):
        """ Reads the dictionary file specified in the config assumed to be stored in data subdirectory """
        directory_path = pkg_resources.resource_filename('pygme', 'data/')
        full_path = os.path.join(directory_path, self.dictionary_filename)
        with open(full_path, "r") as f:
            words = f.readlines()
        self.words = [Word(word.strip("\n")) for word in words]

    def get_random_word(self, min_length: int = 1, max_length: int = sys.maxsize) -> str:
        """ Retrieves a random word with length between the given minimum and max length arguments

        :param min_length - the minimum length that the word should have
        :param max_length - the maximum length that the word should have

        :returns a random word from the dictionary matching the given length criteria
        """
        eligible_words = [word for word in self.words if min_length <= len(word) <= max_length]
        return random.choice(eligible_words)



import random


def get_coordinates_between_limits(grid_width: int, grid_length: int, exclusion_set: set = None) -> tuple:
    """ Provides random coordinates between the specified limits on a 2D grid

    :param grid_width - the width of the grid in number of squares
    :param grid_length - the length of the grid in number of squares
    :param exclusion_set - an optional set of coordinates to include from the population of coordinates to sample
    :returns a tuple with (x-coordinate, y-coordinate)
    """
    # No coordinates to exclude so assume the entire grid is available to sample from
    if not exclusion_set:
        random_x_coordinate = random.randint(0, grid_length - 1)
        random_y_coordinate = random.randint(0, grid_width - 1)
        return random_x_coordinate, random_y_coordinate
    # Only sample from available coordinates on the grid (excluding the given exclusion set of coordinates)
    sample_set = []
    for length_element in range(grid_length):
        for width_element in range(grid_width):
            coordinate = (length_element, width_element)
            if coordinate not in exclusion_set:
                sample_set.append(coordinate)
    if not sample_set:
        return tuple()
    return random.choice(sample_set)


def are_coordinates_between_limits(coordinates: tuple, grid_width: int, grid_length: int) -> bool:
    """ Checks whether the given (x-coordinate, y-coordinate) square is within the given grid dimensions

    :param coordinates - (x-coordinate, y-coordinate) coordinates representing a square on a grid
    :param grid_width - the width of the grid in number of squares
    :param grid_length - the length of the grid in number of squares
    :returns True if the coordinates are within the grid dimensions, False otherwise
    """
    if not coordinates:
        return False
    x_coordinate, y_coordinate = int(coordinates[0]), int(coordinates[1])
    if (0 <= y_coordinate < grid_width) and (0 <= x_coordinate < grid_length):
        return True
    return False


def get_random_directions(possible_directions: list, number_of_directions: int) -> list:
    """ Generates random directions from the possible directions given

    :param possible_directions - a list of possible directions to choose from
    :param number_of_directions - the number of random directions to generate
    :returns a list of [number_of_directions] random directions
    """
    directions = []
    # Make sure possible directions are unique
    possible_directions = list(set(possible_directions))
    possible_directions_count = len(possible_directions)
    for _ in range(number_of_directions):
        random_direction = possible_directions[random.randint(0, possible_directions_count - 1)]
        directions.append(random_direction)
    return directions


def are_contiguous_coordinates(coordinates: list) -> bool:
    """ Checks if the given list of coordinate tuples are in a contiguous sequence of each other in either the X or
    Y direction (diagonals not considered contiguous)

    Example: (1,0), (2,0), (3,0) are contiguous
    (1,0), (2,1), (4,0) are not

    :param coordinates - a list of coordinate tuples (x-coordinate, y-coordinate)
    :returns True if the coordinates are contiguous, false otherwise
    """
    y_coordinates = set()
    x_coordinates = set()
    coordinate_length = len(coordinates)
    for coordinate_tuple in coordinates:
        y_coordinates.add(int(coordinate_tuple[1]))
        x_coordinates.add(int(coordinate_tuple[0]))
    # To be contiguous in either x or y direction, at least one axis needs to be constant (1 unique value only)
    if len(y_coordinates) > 1 and len(x_coordinates) > 1:
        return False
    # Otherwise the difference between the max coordinate value and the minimum must have max increment of 1
    return (max(y_coordinates) - min(y_coordinates) + 1 == coordinate_length or
            max(x_coordinates) - min(x_coordinates) + 1 == coordinate_length)


def get_contiguous_coordinates(starting_coordinate: tuple, direction: str, size: int) -> list:
    """ Provides a list of contiguous coordinate tuples from the starting coordinate in the given direction
    Contiguous coordinates are those in a sequence without skipping in either the X or Y direction

    Example: (1,0), (2,0), (3,0) are contiguous
    (1,0), (2,1), (4,0) are not

    :param starting_coordinate - an (x-coordinate, y-coordinate) tuple of the starting square
    :param direction - one of four directions (right, down, left, up) to generate the sequence in
    :param size - the size of the sequence in number of squares
    :returns a list of coordinate tuples in a contiguous sequence
    """
    assert direction in {"right", "down", "left", "up"}
    coordinate_list = [(int(starting_coordinate[0]), int(starting_coordinate[1]))]
    for idx in range(size - 1):
        if direction == "right":
            new_coordinate = (coordinate_list[idx][0] + 1, coordinate_list[idx][1])
        elif direction == "down":
            new_coordinate = (coordinate_list[idx][0], coordinate_list[idx][1] + 1)
        elif direction == "left":
            new_coordinate = (coordinate_list[idx][0] - 1, coordinate_list[idx][1])
        else:
            new_coordinate = (coordinate_list[idx][0], coordinate_list[idx][1] - 1)
        coordinate_list.append(new_coordinate)
    return coordinate_list


def get_adjacent_coordinates(coordinate: tuple, grid_width: int, grid_length: int) -> list:
    """ Provides a list of the 4 adjacent coordinates around a given coordinate on a 2D grid while skipping any
    coordinates that are out of bounds

    :param coordinate - the coordinate to get the adjacent coordinates for
    :param grid_width - the width of the grid
    :param grid_length - the length of the grid
    :returns a list of tuples containing a max of 4 adjacent coordinates around the given coordinate
    """
    # The given coordinate must be within limits of the grid size
    adjacent_coordinates = []
    if not coordinate:
        return adjacent_coordinates
    assert are_coordinates_between_limits(coordinate, grid_width, grid_length)
    for direction in ["right", "down", "left", "up"]:
        # Get the last coordinate in a list of continuous coordinates including the starting coordinate as first
        adjacent_coordinate = get_contiguous_coordinates(coordinate, direction, 2)[1]
        # Only keep track of adjacent coordinates if they're within the grid boundary
        if are_coordinates_between_limits(adjacent_coordinate, grid_width, grid_length):
            adjacent_coordinates.append(adjacent_coordinate)
    return adjacent_coordinates



from abc import ABC, abstractmethod
import random

from pygme.game.board import GameBoard
from pygme.utils.display import clear_console
from pygme.utils.validation import validate_user_input


class Game(ABC):

    DIFFICULTY_TYPES = {"easy", "normal", "hard"}

    def __init__(self, name: str, config: dict, difficulty: str = "normal") -> None:
        self.name = name
        self.number_of_players = config["number_of_players"]
        self.difficulty = difficulty
        self.players = []
        self.human_player = None
        self.player_turn = 0
        self.config = config
        self.required_inputs = {key: eval(key_type) for key, key_type in config["required_inputs"].items()}

    def _assign_human_player(self) -> None:
        """ Makes a random player in the list of players the human player """
        # Get random index based on length of player list
        random_index = random.randint(0, len(self.players) - 1)
        # Assign a random player in that list to be the human player
        # This player will require input from the current user while the others will play on their own
        self.players[random_index].computer = False
        self.human_player = self.players[random_index]

    def _get_user_input(self, initialization_object: dict = None) -> dict:
        """ Retrieves user input to begin a game

        :param initialization_object - some dictionary containing parameters to initialize with
        :returns the given initialization object or a newly created one if it's not provided
        """
        if not initialization_object:
            initialization_object = {}
            pre_prompt = ""
            while True:
                clear_console()
                try:
                    print("{0}Provide your inputs to begin your game. Difficulty levels: easy, normal, hard\n"
                          .format(pre_prompt))
                    for required_input, input_type in self.required_inputs.items():
                        input_val = input("Enter a value for {0}: ".format(required_input))
                        initialization_object[required_input] = validate_user_input(
                            required_input, input_val, input_type)
                    self._validate_initialization(initialization_object)
                    break
                except Exception as e:
                    pre_prompt = str(e) + "\n\n"
                    pass
        # Validate the input passed through the method arguments
        else:
            self._validate_initialization(initialization_object)
        return initialization_object

    def _next_player(self):
        """ Retrieves the next player in line to play

        :returns a player object retrieved from the list of current players
        """
        player_count = len(self.players)
        if player_count == 0:
            raise RuntimeError("There are no active players in the {0} game".format(self.name))
        # If it's the last player then get the first player in the list
        if self.player_turn == len(self.players) - 1:
            self.player_turn = 0
        # Otherwise get the next player in the list
        else:
            self.player_turn += 1
        return self.players[self.player_turn]

    def _other_players(self) -> list:
        """ Retrieves all the other players besides the one that has the current turn

        :returns a list of player objects
        """
        player_count = len(self.players)
        other_players = []
        if player_count == 0:
            raise RuntimeError("There are no active players in the {0} game".format(self.name))
        for idx in range(player_count):
            if idx == self.player_turn:
                continue
            other_players.append(self.players[idx])
        return other_players

    def print_result(self) -> None:
        """ Prints the result of the game to the user """
        print("Losers:\n")
        for player in self.players:
            if not player.winner:
                print(player.player_id)
        print("Winners:\n")
        for player in self.players:
            if player.winner:
                print(player.player_id)

    @staticmethod
    def construct_board(length: int, width: int, board: GameBoard = None) -> GameBoard:
        """ Creates a new board of the given length and width if an existing board is not provided

        :param length - the length of a new board to create
        :param width - the width of a new board to create
        :param board - an existing board to use
        """
        if not board:
            board = GameBoard(length, width)
        return board

    def _validate_base(self, initialization_object: dict) -> None:
        """ Validates the base config parameters required for any game

        :param initialization_object - a dictionary of game parameters to validate
        """
        # Validate completeness of inputs
        for required_input in self.required_inputs:
            if required_input not in initialization_object:
                raise ValueError("{0} is a required input to begin a {1} game".format(required_input, self.name))
        if initialization_object["difficulty"] not in self.DIFFICULTY_TYPES:
            raise ValueError("The game difficulty must be one of {0}".format(self.DIFFICULTY_TYPES))

    @abstractmethod
    def _validate_initialization(self, initialization_object: dict) -> None:
        pass

    @abstractmethod
    def _initialize(self, initialization_object: dict = None) -> None:
        pass

    @abstractmethod
    def _is_game_over(self) -> bool:
        pass

    @abstractmethod
    def run(self, initialization_object: dict) -> dict:
        pass





class Ship(object):
    """ Represents a ship in the game of Battleship

    Constructor arguments:

    :param ship_type - one of the possible ship types in the game
    :param size - the size of the ship in number of grid squares
    :param representation - how the ship will be represented on a Battleship grid
    :param damaged_representation - how damaged ship segments will be represented on a Battleship grid
    """
    def __init__(self, ship_type, size, representation="0", damaged_representation="*") -> None:
        self.ship_type = ship_type
        self.size = size
        self.representation = representation
        self.damaged_representation = damaged_representation
        # Ship starts out without being placed on the grid and no destroyed segments
        self.coordinates = set()
        self.destroyed_coordinates = set()
        self.placed = False
        self.destroyed = False

    def is_destroyed(self) -> bool:
        """ Checks whether the ship has been destroyed and updates state if so

        :returns True if the ship has been destroyed, False otherwise
        """
        self.destroyed = False
        coordinate_length, destroyed_length = len(self.coordinates), len(self.destroyed_coordinates)
        if coordinate_length > 0 and destroyed_length > 0 and len(self.coordinates - self.destroyed_coordinates) == 0:
            self.destroyed = True
        return self.destroyed

    def place_ship(self, coordinates: list) -> None:
        """ Places the current ship by accepting the given coordinates (assumes ship is being placed on a 2D grid)

        :param coordinates - a list of x and y coordinate tuples
        """
        for coordinate in coordinates:
            self.coordinates.add(coordinate)
        self.placed = True
        self.destroyed = False
        self.destroyed_coordinates = set()

    def get_representation(self, coordinate: tuple) -> str:
        """ Returns the ship segment representation at the given coordinate depending on whether that segment is
        destroyed or not

        :param coordinate - a tuple of x-coordinate, y-coordinate for the segment to return
        :returns a character representation of the ship's segment at the given coordinate
        """
        representation = None
        # Segment is destroyed so return the damaged representation
        if coordinate in self.destroyed_coordinates:
            representation = self.damaged_representation
        # Segment is intact so ship is represented normally
        elif coordinate in self.coordinates:
            representation = self.representation
        return representation

    def take_damage(self, coordinate: tuple) -> tuple:
        """ Accepts damage at the given coordinate

        :param coordinate - tuple of x-coordinate, y-coordinate that was attacked
        :returns a tuple containing booleans for whether the attack was successful and if the ship has been destroyed
        """
        successful_hit, destroyed = False, False
        # Take damage only if part of the ship is on the coordinate
        if coordinate in self.coordinates and coordinate not in self.destroyed_coordinates:
            self.destroyed_coordinates.add(coordinate)
            successful_hit, destroyed = True, self.is_destroyed()
        return successful_hit, destroyed

    def __repr__(self) -> str:
        return "{0} of size {1} with coordinates {2}".format(self.ship_type, self.size, self.coordinates)

    def __str__(self) -> str:
        return "{0} of size {1}".format(self.ship_type, self.size)


class ShipFleet(dict):
    """ Keeps and manages a collection of ships in a dictionary where each ship type is the key and values are ship
    objects

    Constructor arguments:

    :param config - a configuration dictionary containing information on ship types, sizes, etc.
    """
    def __init__(self, config: dict) -> None:
        self.destroyed = False
        assert "ship_types" in config and "size_by_type" in config
        ship_types = config.get("ship_types", set())
        size_by_type = config.get("size_by_type", {})
        super().__init__({
            ship_type: Ship(ship_type, size) for ship_type, size in size_by_type.items() if ship_type in ship_types
        })

    def is_destroyed(self) -> bool:
        """ Checks whether the fleet has been destroyed and updates state accordingly

        :returns True if every ship in the fleet has been destroyed, False if at least 1 has not been destroyed
        """
        self.destroyed = True
        for _, ship in self.items():
            ship_destroyed = ship.is_destroyed()
            # If any one ship is still alive then the fleet is also still alive
            if not ship_destroyed:
                self.destroyed = False
        return self.destroyed

    def ship_is_hit(self, coordinate: tuple) -> bool:
        """ Checks whether a ship in the fleet has already been hit at the given coordinate

        :param coordinate - a tuple of x and y coordinates to check for hits
        :returns True if a ship has already been hit, False otherwise
        """
        for _, ship in self.items():
            if coordinate in ship.destroyed_coordinates:
                return True
        return False

    def accept_attack(self, coordinate: tuple) -> tuple:
        """ Accepts an attack on the fleet by transmitting the damage to a ship in the fleet

        :param coordinate - a tuple of x and y coordinates that are being attacked
        :returns whether there has been a successful hit and a ship has been destroyed and the new ship segment repr
        """
        successful_hit, ship_destroyed, representation = False, False, None
        for _, ship in self.items():
            successful_hit, ship_destroyed = ship.take_damage(coordinate)
            if successful_hit:
                representation = ship.get_representation(coordinate)
                break
        return successful_hit, ship_destroyed, representation

    def _print(self, print_function) -> str:
        """ Utility method for repr and str magic methods to print the fleet and each ship in it

        :param print_function - either repr or str to use for representing individual ship objects when printing the
            fleet object
        :returns a string representation of the fleet collection object
        """
        return_str = "Ship fleet: "
        for _, ship in self.items():
            return_str += "{0}, ".format(print_function(ship))
        return return_str[:-2]

    @property
    def unique_ship_representations(self) -> set:
        """ Provides a collection of unique ship representation characters when displayed in the terminal

        :returns a set of representation characters
        """
        ship_representations = set()
        for _, ship in self.items():
            ship_representations.add(ship.representation)
        return ship_representations

    def __repr__(self) -> str:
        return self._print(repr)

    def __str__(self) -> str:
        return self._print(str)
